import numpy
from numpy import pi, sin, cos, arccos, tan
import lightrayrider as raytrace

def cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d):
	t = raytrace.cone_raytrace_finite(thetas, rhos, x, y, z, a, b, c, d)
	t[t == -1] = 1e10
	x += a*t
	y += b*t
	z += c*t

def to_spherical(pos):
	(xf, yf, zf) = pos
	xf = numpy.asarray(xf).reshape((-1,))
	yf = numpy.asarray(yf).reshape((-1,))
	zf = numpy.asarray(zf).reshape((-1,))
	rad = (xf**2+yf**2+zf**2)**0.5
	#phi = numpy.fmod(numpy.arctan2(yf, xf) + 2*pi, pi)
	phi = numpy.arctan2(yf, xf)
	theta = numpy.where(rad == 0, 0., arccos(zf / rad))
	return (rad, theta, phi)

def to_cartesian(pos):
	(rad, theta, phi) = pos
	xv = rad * sin(theta) * cos(phi)
	yv = rad * sin(theta) * sin(phi)
	zv = rad * cos(theta)
	return (xv, yv, zv)

def test_central_single_t():
	thetas = numpy.array([45 * pi / 180.])
	rhos = numpy.array([1.0])
	x = numpy.zeros(3)
	y = numpy.zeros(3)
	z = numpy.zeros(3)
	beta = numpy.array([0, 44, 46]) / 180. * numpy.pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.5, 0.5, 0.5])
	t = raytrace.cone_raytrace_finite(thetas, rhos, x, y, z, a, b, c, d)
	# now x, y, z should be updated
	assert t[0] == -1, r[0]
	assert t[1] == -1, r[1]
	assert t[2] < 1,   r[2]
	assert numpy.isclose(t[2], 0.5),  t[2]

def test_central_single():
	thetas = numpy.array([45 * pi / 180.])
	rhos = numpy.array([1.0])
	x = numpy.zeros(3)
	y = numpy.zeros(3)
	z = numpy.zeros(3)
	beta = numpy.array([0, 44, 46]) / 180. * numpy.pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.5, 0.5, 0.5])
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	# now x, y, z should be updated
	r = (x**2 + y**2 + z**2)**0.5
	print(r, x, y, z)
	assert r[0] > 10, r[0]
	assert r[1] > 10, r[1]
	assert r[2] < 1,  r[2]
	assert numpy.isclose(r[2], 0.5),  r[2]
	
	# test bottom part
	x = numpy.zeros(3)
	y = numpy.zeros(3)
	z = numpy.zeros(3)
	beta = numpy.array([90, 90+44, 90+46]) / 180. * numpy.pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.5, 0.5, 0.5])
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	# now x, y, z should be updated
	r = (x**2 + y**2 + z**2)**0.5
	print(r, x, y, z)
	assert numpy.isclose(r[0], 0.5),  r[0]
	assert numpy.isclose(r[1], 0.5),  r[1]
	assert r[2] > 10, r[2]

	# test depth
	x = numpy.zeros(3)
	y = numpy.zeros(3)
	z = numpy.zeros(3)
	beta = numpy.array([90, 90, 90]) / 180. * numpy.pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.1, 0.9, 1.1])
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	# now x, y, z should be updated
	r = (x**2 + y**2 + z**2)**0.5
	print(r, x, y, z)
	assert numpy.isclose(r[0], 0.1),  r[0]
	assert numpy.isclose(r[1], 0.9),  r[1]
	assert r[2] > 10, r[2]

def test_border_single():
	# test at the border
	thetas = numpy.array([60 * pi / 180.])
	rhos = numpy.array([1.0])
	x = numpy.zeros(3) + 0.99
	y = numpy.zeros(3)
	z = numpy.zeros(3)
	x0, y0, z0 = numpy.copy(x), numpy.copy(y), numpy.copy(z)
	beta = numpy.array([0, 0, 0]) / 180. * numpy.pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.1, 0.9, 1.1])
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	# now x, y, z should be updated
	r = ((x-x0)**2 + y**2 + z**2)**0.5
	print(r, x, y, z)
	assert numpy.isclose(r[0], 0.1),  r[0]
	assert r[1] > 10, r[1]
	assert r[2] > 10, r[2]

def test_crossing_single_center():
	# test near the center
	thetas = numpy.array([45 * pi / 180.])
	rhos = numpy.array([1.0])
	x = numpy.zeros(3) + 0.01
	y = numpy.zeros(3)
	z = numpy.zeros(3) - 1
	x0, y0, z0 = numpy.copy(x), numpy.copy(y), numpy.copy(z)
	beta = numpy.array([0, 0, 0]) / 180. * numpy.pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.01, 0.3, 1.1])
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	# now x, y, z should be updated
	r = ((x-x0)**2 + (y - y0)**2 + (z-z0)**2)**0.5
	print(r, x, y, z)
	assert numpy.isclose(r[0], 1),  r[0]
	assert r[1] > 10, r[1]
	assert r[2] > 10, r[2]

def test_crossing_single_border():
	# test at the border, where sphere plays the more important role
	thetas = numpy.array([45 * pi / 180.])
	rhos = numpy.array([1.0])
	x = numpy.zeros(3) + 0.99
	y = numpy.zeros(3)
	z = numpy.zeros(3) - 1
	x0, y0, z0 = numpy.copy(x), numpy.copy(y), numpy.copy(z)
	beta = numpy.array([0, 0, 0]) / 180. * numpy.pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.1, 0.3, 1.1])
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	# now x, y, z should be updated
	r = ((x-x0)**2 + (y - y0)**2 + (z-z0)**2)**0.5
	print(r, x, y, z)
	assert numpy.isclose(r[0], 0.958932),  r[0]
	assert r[1] > 10, r[1]
	assert r[2] > 10, r[2]
	
def test_noncentral_single():
	# almost central case, result should be the same
	thetas = numpy.array([45 * pi / 180.])
	rhos = numpy.array([1.0])
	x = numpy.zeros(3) + 0.001
	y = numpy.zeros(3)
	z = numpy.zeros(3)
	beta = numpy.array([0, 44, 45]) / 180. * pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.5, 0.5, 0.5])
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	# now x, y, z should be updated
	r = ((x - 0.001)**2 + y**2 + z**2)**0.5
	print(r, x, y, z)
	assert r[0] > 10, r[0]
	assert r[1] > 10, r[1]
	assert r[2] < 1,  r[2]
	assert numpy.isclose(r[2], 0.5),  r[2]
	
def test_single_horizontal():
	# cross the sphere and cone
	thetas = numpy.array([45 * pi / 180.])
	rhos = numpy.array([1.0])
	x = numpy.zeros(3) - 2
	y = numpy.zeros(3)
	z = numpy.zeros(3) + 0.001
	a = numpy.ones(3)
	b = numpy.zeros(3)
	c = numpy.zeros(3)
	d = numpy.array([0.001, 1., 1.9])
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	print(x, y, z)
	xexp = [-1 + 0.001, 0.001*2, 1+0.001*2-0.1]
	assert numpy.allclose(x, xexp, atol=0.001), (x, xexp, x-xexp)
	assert numpy.allclose(y, [0, 0, 0]), y
	assert numpy.allclose(z, [0.001, 0.001, 0.001]), z


def test_central_multiple():
	thetas = numpy.array([45, 60]) * pi / 180.
	rhos = numpy.array([1.0, 2.0])
	x = numpy.zeros(3)
	y = numpy.zeros(3)
	z = numpy.zeros(3)
	beta = numpy.array([44, 46, 70]) / 180. * numpy.pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.5, 0.5, 0.5])
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	# now x, y, z should be updated
	r = (x**2 + y**2 + z**2)**0.5
	print(r, x, y, z)
	assert r[0] > 10, r[0]
	assert r[1] < 1, r[1]
	assert numpy.isclose(r[1], 0.5),  r[2]
	assert r[2] < 1, r[2]
	assert numpy.isclose(r[2], 0.25),  r[2]
	
	# test bottom part
	x = numpy.zeros(3)
	y = numpy.zeros(3)
	z = numpy.zeros(3)
	beta = numpy.array([90+20, 90+44, 90+46]) / 180. * numpy.pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.5, 0.5, 0.5])
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	# now x, y, z should be updated
	r = (x**2 + y**2 + z**2)**0.5
	print(r, x, y, z)
	assert numpy.isclose(r[0], 0.25),  r[0]
	assert numpy.isclose(r[1], 0.5),  r[1]
	assert r[2] > 10, r[2]

	# test depth
	x = numpy.zeros(3)
	y = numpy.zeros(3)
	z = numpy.zeros(3)
	beta = numpy.array([90, 90, 90]) / 180. * numpy.pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.1, 0.9, 1.1]) * 2
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	# now x, y, z should be updated
	r = (x**2 + y**2 + z**2)**0.5
	print(r, x, y, z)
	assert numpy.isclose(r[0], 0.1),  r[0]
	assert numpy.isclose(r[1], 0.9),  r[1]
	assert r[2] > 10, r[2]

def test_speed_central_multiple():
	thetas = numpy.linspace(30, 90, 30) * pi / 180.
	rhos = 10**numpy.linspace(0.1, 2.0, 30)
	N = 1000000
	x = numpy.zeros(N)
	y = numpy.zeros(N)
	z = numpy.zeros(N)
	beta = numpy.random.uniform(0, 90, size=N) / 180. * numpy.pi
	phi = numpy.random.uniform(0, 2*pi, size=N)
	a, b, c = to_cartesian((1, beta, phi))
	d = 10**numpy.random.uniform(-2, 1, size=N) + 0.001
	cone_raytrace_finite_coords(thetas, rhos, x, y, z, a, b, c, d)
	print(x, y, z)
	assert not (x == 0).any()
	assert not (y == 0).any()
	assert not (z == 0).any()

"""
def test_single():
	xx = numpy.array([0.])
	yy = numpy.array([1.])
	zz = numpy.array([0.])
	RR = numpy.array([0.1])
	rho = numpy.array([5.])
	a = numpy.array([1,0,0.])
	b = numpy.array([0,1,0.])
	c = numpy.array([0,0,1.])
	x = numpy.zeros(3)
	y = x
	z = x
	NHmax = numpy.array([0.5,0.5,0.5])
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	print t
	assert numpy.allclose(t, [-1, 1, -1])
	
	a = numpy.array([0,0,0.])
	b = numpy.array([1,1,1.])
	c = numpy.array([0,0,0.])
	NHmax = numpy.array([1e-10,1 - 1e-10,2])
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	print t
	assert numpy.allclose(t, [0.9, 1.1, -1])

def test_multiple():
	xx = numpy.array([0.,0])
	yy = numpy.array([1.,2])
	zz = numpy.array([0.,0])
	RR = numpy.array([0.1, 0.1])
	rho = numpy.array([5., 5.])
	a = numpy.array([0,0,0.])
	b = numpy.array([1,1,1.])
	c = numpy.array([0,0,0.])
	x = numpy.zeros(3)
	y = x
	z = x
	NHmax = numpy.array([0.999999,1.000001,1.999999])
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	print t
	assert numpy.allclose(t, [1.1, 1.9, 2.1])

def test_multiple_diagonal():
	s3 = numpy.sqrt(3)
	xx = numpy.array([-3,-2,-1,1,2,3.])
	yy = numpy.array([-3,-2,-1,1,2,3.])
	zz = numpy.array([-3,-2,-1,1,2,3.])
	print 'xx', xx
	RR = numpy.array([0.1, 0.1, 0.1, 0.1, 0.1, 0.1])
	rho = numpy.array([5., 5., 5., 5., 5., 5.])
	a = numpy.array([1,1,1.,1]) / s3
	b = numpy.array([1,1,1.,1]) / s3
	c = numpy.array([1,1,1.,1]) / s3
	x = numpy.zeros(4)
	y = x
	z = x
	NHmax = numpy.array([0.99999, 1.99999, 2.9999, 3.01])
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	print t
	assert numpy.allclose(t, [s3+0.1, s3*2+0.1, s3*3+0.1, -1])
	NHmax = numpy.array([0.00001, 1.00001, 2.00001, 3.0001])
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	print t
	assert numpy.allclose(t, [s3-0.1, s3*2-0.1, s3*3-0.1, -1])

def test_angles():
	xx = numpy.array([0.])
	yy = numpy.array([1.])
	zz = numpy.array([0.])
	RR = numpy.array([0.1])
	rho = numpy.array([5.])
	
	phi = numpy.linspace(0, 2*numpy.pi, 400)
	a = numpy.sin(phi)
	b = numpy.cos(phi)
	c = b * 0
	x = numpy.zeros(len(a))
	y = x
	z = x
	NHmax = c * 0 + 0.5
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	assert t[0] == 1
	assert t[1] > 1
	assert t[10] == -1
	assert t[-10] == -1
	assert t[-2] > 1
	assert t[-1] == 1
	
"""



import numpy
import lightrayrider as raytrace

def to_cartesian(pos):
	(rad, theta, phi) = pos
	sin, cos = numpy.sin, numpy.cos
	xv = rad * sin(theta) * cos(phi)
	yv = rad * sin(theta) * sin(phi)
	zv = rad * cos(theta)
	return (xv, yv, zv)

def test_single():
	xx = numpy.array([0.])
	yy = numpy.array([1.])
	zz = numpy.array([0.])
	RR = numpy.array([0.1])
	rho = numpy.array([5.])
	a = numpy.array([1,0,0.])
	b = numpy.array([0,1,0.])
	c = numpy.array([0,0,1.])
	x = numpy.zeros(3)
	y = x
	z = x
	NHmax = numpy.array([0.5,0.5,0.5])
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	print(t)
	assert numpy.allclose(t, [-1, 1, -1])
	
	a = numpy.array([0,0,0.])
	b = numpy.array([1,1,1.])
	c = numpy.array([0,0,0.])
	NHmax = numpy.array([1e-10,1 - 1e-10,2])
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	print(t)
	assert numpy.allclose(t, [0.9, 1.1, -1])

def test_crossing_single():
	xx = numpy.array([0.])
	yy = numpy.array([0.])
	zz = numpy.array([0.])
	RR = numpy.array([1.])
	rho = numpy.array([1.])
	# test at the border
	x = numpy.zeros(3) + 0.99
	y = numpy.zeros(3)
	z = numpy.zeros(3) - 1
	x0, y0, z0 = numpy.copy(x), numpy.copy(y), numpy.copy(z)
	beta = numpy.array([0, 0, 0]) / 180. * numpy.pi
	a, b, c = to_cartesian((1, beta, 0))
	d = numpy.array([0.1, 0.3, 1.1])
	r = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, d)
	print(r)
	assert numpy.isclose(r[0], 0.958932),  r[0]
	assert r[1] == -1, r[1]
	assert r[2] == -1, r[2]

def test_multiple():
	xx = numpy.array([0.,0])
	yy = numpy.array([1.,2])
	zz = numpy.array([0.,0])
	RR = numpy.array([0.1, 0.1])
	rho = numpy.array([5., 5.])
	a = numpy.array([0,0,0.])
	b = numpy.array([1,1,1.])
	c = numpy.array([0,0,0.])
	x = numpy.zeros(3)
	y = x
	z = x
	NHmax = numpy.array([0.999999,1.000001,1.999999])
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	print(t)
	assert numpy.allclose(t, [1.1, 1.9, 2.1])

def test_multiple_diagonal():
	s3 = numpy.sqrt(3)
	xx = numpy.array([-3,-2,-1,1,2,3.])
	yy = numpy.array([-3,-2,-1,1,2,3.])
	zz = numpy.array([-3,-2,-1,1,2,3.])
	print('xx', xx)
	RR = numpy.array([0.1, 0.1, 0.1, 0.1, 0.1, 0.1])
	rho = numpy.array([5., 5., 5., 5., 5., 5.])
	a = numpy.array([1,1,1.,1]) / s3
	b = numpy.array([1,1,1.,1]) / s3
	c = numpy.array([1,1,1.,1]) / s3
	x = numpy.zeros(4)
	y = x
	z = x
	NHmax = numpy.array([0.99999, 1.99999, 2.9999, 3.01])
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	print(t)
	assert numpy.allclose(t, [s3+0.1, s3*2+0.1, s3*3+0.1, -1])
	NHmax = numpy.array([0.00001, 1.00001, 2.00001, 3.0001])
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	print(t)
	assert numpy.allclose(t, [s3-0.1, s3*2-0.1, s3*3-0.1, -1])

def test_angles():
	xx = numpy.array([0.])
	yy = numpy.array([1.])
	zz = numpy.array([0.])
	RR = numpy.array([0.1])
	rho = numpy.array([5.])
	
	phi = numpy.linspace(0, 2*numpy.pi, 400)
	a = numpy.sin(phi)
	b = numpy.cos(phi)
	c = b * 0
	x = numpy.zeros(len(a))
	y = x
	z = x
	NHmax = c * 0 + 0.5
	t = raytrace.sphere_raytrace_finite(xx, yy, zz, RR, rho, x, y, z, a, b, c, NHmax)
	assert t[0] == 1
	assert t[1] > 1
	assert t[10] == -1
	assert t[-10] == -1
	assert t[-2] > 1
	assert t[-1] == 1
	




import os
import numpy
import scipy.interpolate
from lightrayrider import voronoi_raytrace

def montecarlo(x, y, z, R, rho, a, b, c, mindistances):
	r = (x**2 + y**2 + z**2)**0.5
	rmax = r.max()
	interpolator = scipy.interpolate.NearestNDInterpolator(numpy.transpose([x, y, z]), rho)
	N = 400000
	t = numpy.logspace(-3, numpy.log10(rmax), N)
	#t = numpy.linspace(0, r.min(), N)
	#t = numpy.hstack((numpy.linspace(0, r.min(), 10)[:-1], numpy.logspace(numpy.log10(r.min()), numpy.log10(rmax), N/40)))
	#t = numpy.array(sorted(r))
	points = numpy.transpose([a * t, b * t, c * t])
	densities = interpolator(points)
	#print densities
	#print densities.max(), densities.min()
	#for ti, rhoi in zip(t, densities):
	#	print('  %.3e %.3f' % (rhoi, ti))
	#plt.plot(ti, rhoi, '-')
	#print rho.max(), rho.min()
	NH = [numpy.trapz(x=t, y=densities[t >= mindistance]) for mindistance in mindistances]
	print('0...%.2e' % rmax, 'NH:', zip(mindistances, NH))
	return numpy.array(NH)

def test():
	data_orig = numpy.load(os.path.join(os.path.dirname(__file__), 'ray_example.npz'))
	data = dict([(k, data_orig[k].astype(numpy.float64)) for k in data_orig.keys()])
	x = data['x']
	y = data['y']
	z = data['z']
	R = data['R']
	rho = data['conversion'][:,0]
	mindistances = numpy.copy(data['mindistances'][:1])
	print('densities:', rho)
	
	for i in range(3):
		vall = data['v'][i:i+1,:]
		a, b, c = numpy.copy(vall[:,0]), numpy.copy(vall[:,1]), numpy.copy(vall[:,2])
		
		print('MONTE CARLO:', i)
		result = montecarlo(x, y, z, R, rho, a, b, c, mindistances)
		print(result)
	
		#print('SEGMENTATION:')
		#points = numpy.transpose([x, y, z])
		#segments = line.segment(vall[0,:], points)
		#for xlo, xhi, i in segments:
		#	print(xlo, xhi, rho[i])
		#return result
	
		for i in range(1):
			print('running...', i)
			result2 = voronoi_raytrace(x, y, z, R, rho, a, b, c, mindistances).flatten()
		print('done.')
		print(result)
		assert result.shape == result2.shape, (result.shape, result2.shape)
		print('result:', numpy.log10(result))
		print('reference:', numpy.log10(result2))
		print('absdiff:', numpy.max(numpy.abs(result - result2)))
		print('reldiff:', numpy.max(numpy.abs((result - result2)/result2)))
		assert numpy.allclose(result, result2, rtol=0.5, atol=1e20)
	

if __name__ == '__main__':
	test()




import numpy
from math import floor, ceil
from lightrayrider import grid_raytrace_finite

def raytrace_grid_finite(x, v, d):
	(x0, y0, z0), (dx, dy, dz) = x, v
	# call raytrace_grid_finite_c()
	rho = numpy.ones((256, 256, 256))
	x, y, z = numpy.array([x0*1. + 128]), numpy.array([y0*1.+128]), numpy.array([z0*1.+128])
	a, b, c = numpy.array([dx*1.]), numpy.array([dy*1.]), numpy.array([dz*1.])
	NHmax = numpy.array([d*1.])
	r = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	t = r[0]
	print('distance:', t, (x, y, z), NHmax)
	return x0 + dx * t, y0 + dy * t, z0 + dz * t
	
def test_smallstep():
	x, y, z = numpy.zeros(3)
	dx, dy, dz = 0.99, 0.01, 0
	d = 0.5
	(x1, y1, z1) = raytrace_grid_finite((x, y, z), (dx, dy, dz), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, 0.5*0.99), (x1, 0.5*0.99) 
	assert numpy.allclose(y1, 0.5*0.01), (y1, 0.5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)

	x, y, z = numpy.zeros(3)
	dx, dy, dz = 0.99, 0.01, 0
	d = 0.5
	(y1, x1, z1) = raytrace_grid_finite((y, x, z), (dy, dx, dz), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, 0.5*0.99), (x1, 0.5*0.99) 
	assert numpy.allclose(y1, 0.5*0.01), (y1, 0.5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)

	x, y, z = numpy.zeros(3)
	dx, dy, dz = 0.99, 0.01, 0
	d = 0.5
	(z1, y1, x1) = raytrace_grid_finite((z, y, x), (dz, dy, dx), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, 0.5*0.99), (x1, 0.5*0.99) 
	assert numpy.allclose(y1, 0.5*0.01), (y1, 0.5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)

	x, y, z = numpy.zeros(3)
	dx, dy, dz = 0.99, 0.01, 0
	d = 0.5
	(z1, x1, y1) = raytrace_grid_finite((z, x, y), (dz, dx, dy), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, 0.5*0.99), (x1, 0.5*0.99) 
	assert numpy.allclose(y1, 0.5*0.01), (y1, 0.5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)

def test_smallstep_offset():
	x, y, z = numpy.ones(3)
	dx, dy, dz = 0.99, 0.01, 0
	d = 0.5
	(x1, y1, z1) = raytrace_grid_finite((x, y, z), (dx, dy, dz), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, 1+0.5*0.99), (x1, 1+0.5*0.99) 
	assert numpy.allclose(y1, 1+0.5*0.01), (y1, 1+0.5*0.01) 
	assert numpy.allclose(z1, 1+0), (z1, 1+0)

	x, y, z = numpy.ones(3)
	dx, dy, dz = 0.99, 0.01, 0
	d = 0.5
	(y1, x1, z1) = raytrace_grid_finite((y, x, z), (dy, dx, dz), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, 1+0.5*0.99), (x1, 1+0.5*0.99) 
	assert numpy.allclose(y1, 1+0.5*0.01), (y1, 1+0.5*0.01) 
	assert numpy.allclose(z1, 1+0), (z1, 1+0)

	x, y, z = numpy.ones(3)
	dx, dy, dz = 0.99, 0.01, 0
	d = 0.5
	(z1, y1, x1) = raytrace_grid_finite((z, y, x), (dz, dy, dx), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, 1+0.5*0.99), (x1, 1+0.5*0.99) 
	assert numpy.allclose(y1, 1+0.5*0.01), (y1, 1+0.5*0.01) 
	assert numpy.allclose(z1, 1+0), (z1, 1+0)


def test_smallstep_negative():
	x, y, z = numpy.zeros(3)
	dx, dy, dz = -0.99, -0.01, 0
	d = 0.5
	(x1, y1, z1) = raytrace_grid_finite((x, y, z), (dx, dy, dz), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, -0.5*0.99), (x1, -0.5*0.99) 
	assert numpy.allclose(y1, -0.5*0.01), (y1, -0.5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)

	x, y, z = numpy.zeros(3)
	dx, dy, dz = -0.99, -0.01, 0
	d = 0.5
	(y1, x1, z1) = raytrace_grid_finite((y, x, z), (dy, dx, dz), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, -0.5*0.99), (x1, -0.5*0.99) 
	assert numpy.allclose(y1, -0.5*0.01), (y1, -0.5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)

	x, y, z = numpy.zeros(3)
	dx, dy, dz = -0.99, -0.01, 0
	d = 0.5
	(z1, y1, x1) = raytrace_grid_finite((z, y, x), (dz, dy, dx), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, -0.5*0.99), (x1, -0.5*0.99) 
	assert numpy.allclose(y1, -0.5*0.01), (y1, -0.5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)


def test_negative():
	x, y, z = numpy.zeros(3)
	dx, dy, dz = -0.99, -0.01, 0
	d = 5
	(x1, y1, z1) = raytrace_grid_finite((x, y, z), (dx, dy, dz), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, -5*0.99), (x1, -5*0.99) 
	assert numpy.allclose(y1, -5*0.01), (y1, -5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)

	x, y, z = numpy.zeros(3)
	dx, dy, dz = -0.99, -0.01, 0
	d = 5
	(y1, x1, z1) = raytrace_grid_finite((y, x, z), (dy, dx, dz), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, -5*0.99), (x1, -5*0.99) 
	assert numpy.allclose(y1, -5*0.01), (y1, -5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)

	x, y, z = numpy.zeros(3)
	dx, dy, dz = -0.99, -0.01, 0
	d = 5
	(z1, y1, x1) = raytrace_grid_finite((z, y, x), (dz, dy, dx), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, -5*0.99), (x1, -5*0.99)
	assert numpy.allclose(y1, -5*0.01), (y1, -5*0.01)
	assert numpy.allclose(z1, 0), (z1, 0)

def test_reverse():
	# go back
	start = numpy.zeros(3)
	dx, dy, dz = 0.99, 0.01, 0
	d = 0.5
	mid = raytrace_grid_finite(start, (dx, dy, dz), d)
	dx *= -1
	dy *= -1
	dz *= -1
	final = raytrace_grid_finite(mid, (dx, dy, dz), d)
	# should land at 5, 0, 0
	print(start, final)
	assert numpy.allclose(start, final), (start, final) 


	dx, dy, dz = 0.99, 0.01, 0
	d = 1.5
	mid = raytrace_grid_finite(start, (dx, dy, dz), d)
	dx *= -1
	dy *= -1
	dz *= -1
	final = raytrace_grid_finite(mid, (dx, dy, dz), d)
	# should land at 5, 0, 0
	print(start, final)
	assert numpy.allclose(start, final), (start, final) 

	start = numpy.zeros(3)
	dx, dy, dz = 0.5, 0.5, 0.5
	d = 0.5
	mid = raytrace_grid_finite(start, (dx, dy, dz), d)
	dx *= -1
	dy *= -1
	dz *= -1
	final = raytrace_grid_finite(mid, (dx, dy, dz), d)
	# should land at 5, 0, 0
	assert numpy.allclose(start, final), (start, final) 
	

def test_straight():
	x, y, z = numpy.zeros(3)
	dx, dy, dz = 0.99, 0.01, 0
	d = 5.5
	(x1, y1, z1) = raytrace_grid_finite((x, y, z), (dx, dy, dz), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, 5.5*0.99), (x1, 5.5*0.99) 
	assert numpy.allclose(y1, 5.5*0.01), (y1, 5.5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)

	x, y, z = numpy.zeros(3)
	dx, dy, dz = 0.99, 0.01, 0
	d = 5.5
	(y1, x1, z1) = raytrace_grid_finite((y, x, z), (dy, dx, dz), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, 5.5*0.99), (x1, 5.5*0.99) 
	assert numpy.allclose(y1, 5.5*0.01), (y1, 5.5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)

	x, y, z = numpy.zeros(3)
	dx, dy, dz = 0.99, 0.01, 0
	d = 5.5
	(z1, y1, x1) = raytrace_grid_finite((z, y, x), (dz, dy, dx), d)
	# should land at 5, 0, 0
	assert numpy.allclose(x1, 5.5*0.99), (x1, 5.5*0.99) 
	assert numpy.allclose(y1, 5.5*0.01), (y1, 5.5*0.01) 
	assert numpy.allclose(z1, 0), (z1, 0)


def test_diagonal():
	x, y, z = numpy.zeros(3)
	dx, dy, dz = 2**0.5, 2**0.5, 0
	d = 5.5
	(x1, y1, z1) = raytrace_grid_finite((x, y, z), (dx, dy, dz), d)
	# should land at 2, 2, 0
	assert numpy.allclose(x1, 5.5*dx), (x1, 5.5*dx)
	assert numpy.allclose(y1, 5.5*dy), (y1, 5.5*dy)
	assert numpy.allclose(z1, 0), (z1, 0)

def test_random():
	for i in range(40):
		x0 = numpy.random.normal(0, 3, size=3)
		dv = numpy.random.normal(size=3)
		dv /= (dv**2).sum()
		d = numpy.random.normal(0, 3)**2
		# expected
		final = x0 + dv * d
		print(x0, dv, d, '--> expect:', final)
		x1 = raytrace_grid_finite(x0, dv, d)
		assert numpy.allclose(final, x1), (final, x1)

def test_random_single():
	x, y, z = numpy.array([125.275611]), numpy.array([127.322529]), numpy.array([128.700102])
	a, b, c = numpy.array([0.22951647]), numpy.array([-0.11787839]), numpy.array([0.19780523])
	print('direction:', a**2 + b**2 + c**2)
	NHmax = numpy.array([63.1791943441])
	rho = numpy.ones((256, 256, 256))
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	print('distance:', t)
	assert (t > 0).all(), t
	assert numpy.isclose(t, NHmax), (t, NHmax)
	xfinal = x + a * NHmax
	yfinal = y + b * NHmax
	zfinal = z + c * NHmax

def test_single_small1d_pos():
	x, y, z = numpy.array([128.]), numpy.array([128.]), numpy.array([128.])
	a, b, c = numpy.array([0.687837]), numpy.array([0.725799]), numpy.array([0.009826])
	print('direction:', a**2 + b**2 + c**2)
	NHmax = numpy.array([1000.])
	rho = numpy.ones((256, 256, 256))
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	print('distance:', t)
	assert t == -1, t
	NHmax = numpy.array([10.])
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	assert (t > 0).all(), t
	assert numpy.isclose(t, NHmax), (t, NHmax)
	
def test_single_small1d_neg():
	x, y, z = numpy.array([128.]), numpy.array([128.]), numpy.array([128.])
	a, b, c = numpy.array([-0.687837]), numpy.array([-0.725799]), numpy.array([0.009826])
	print('direction:', a**2 + b**2 + c**2)
	NHmax = numpy.array([1000.])
	rho = numpy.ones((256, 256, 256))
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	print('distance:', t)
	assert t == -1, t
	NHmax = numpy.array([10.])
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	assert (t > 0).all(), t
	assert numpy.isclose(t, NHmax), (t, NHmax)
	
def test_inhomogeneous_single():
	rho = numpy.zeros((256, 256, 256))
	rho[127:130,127:130,127:130] = 1
	rho[128,128,128] = 10
	
	x, y, z = [numpy.array([128.5])]*3
	a, b, c = numpy.array([-0.321689]), numpy.array([-0.901964]), numpy.array([-0.288056])
	print('direction:', a**2 + b**2 + c**2)
	NHmax = numpy.array([(10 + 1 + 1) * 3**0.5])
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	print('distance:', t)
	assert (t == -1).all(), t


def test_density_coord():
	rho = numpy.zeros((256, 256, 256))
	rho[128,128,128:140] = 1
	x, y, z = [numpy.array([128.]*6)]*3
	a = numpy.array([1., 0., 0., -1., 0., 0.])
	b = numpy.array([0., 1., 0., 0., -1., 0.])
	c = numpy.array([0., 0., 1., 0., 0., -1.])
	NHmax = numpy.array([3.]*6)
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	t_expect = numpy.array([-1,-1,3., -1,-1,-1])
	assert numpy.allclose(t, t_expect), (t, t_expect)

	rho = numpy.zeros((256, 256, 256))
	rho[128,128:140,128] = 1
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	t_expect = numpy.array([-1,3.,-1, -1,-1,-1])
	assert numpy.allclose(t, t_expect), (t, t_expect)

	rho = numpy.zeros((256, 256, 256))
	rho[128:140,128,128] = 1
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	t_expect = numpy.array([3.,-1,-1, -1,-1,-1])
	assert numpy.allclose(t, t_expect), (t, t_expect)

	rho = numpy.zeros((256, 256, 256))
	rho[128,128,120:129] = 1
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	t_expect = numpy.array([-1,-1,-1, -1,-1,3.])
	assert numpy.allclose(t, t_expect), (t, t_expect)
	
	rho = numpy.zeros((256, 256, 256))
	rho[128,120:129,128] = 1
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	t_expect = numpy.array([-1,-1,-1, -1,3.,-1])
	assert numpy.allclose(t, t_expect), (t, t_expect)

	rho = numpy.zeros((256, 256, 256))
	rho[120:129,128,128] = 1
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	t_expect = numpy.array([-1,-1,-1, 3.,-1,-1])
	assert numpy.allclose(t, t_expect), (t, t_expect)



def test_inhomogeneous():
	rho = numpy.zeros((256, 256, 256))
	rho[128,128,128] = 10
	N = 40
	#rho[124:132,124:132,124:132] = 1
	x, y, z = [numpy.zeros(N) + 128.5]*3
	dv = numpy.random.normal(size=(N, 3))
	dv /= (dv**2).sum(axis=1).reshape((-1, 1))**0.5
	a, b, c = dv.transpose()
	a, b, c = numpy.array(a), numpy.array(b), numpy.array(c)
	# call raytrace_grid_finite_c()
	NHmax = 4.99 + numpy.zeros((N))
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	print('distance:', t)
	assert numpy.allclose(t, 0.499), t
	
	# go one further
	rho[124:132,124:132,124:132] = 1
	rho[128,128,128] = 10
	NHmax = 5.99 + numpy.zeros((N))
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	print('distance:', t)
	assert (t <= 1.499).all(), t
	assert (t >= 0.599).all(), t
	
	# go even further
	rho = numpy.zeros((256, 256, 256))
	rho[127:130,127:130,127:130] = 1
	rho[128,128,128] = 10
	NHmax = (10 + 1 + 1) * 3**0.5 + numpy.zeros((N))
	#NHmax = 400 + numpy.zeros((20))
	t = grid_raytrace_finite(rho, x, y, z, a, b, c, NHmax)
	print('distance:', t)
	assert (t == -1).all(), t
	

	




import numpy
import os
from lightrayrider import sphere_raytrace

def rootterm(a, b, c, x, y, z, R):
	return (R**2*a**2 + R**2*b**2 + R**2*c**2 - a**2*y**2 - a**2*z**2 + 2*a*b*x*y + 2*a*c*x*z - b**2*x**2 - b**2*z**2 + 2*b*c*y*z - c**2*x**2 - c**2*y**2)

def cross_neg(a, b, c, x, y, z, R, rootterm):
	return (a*x + b*y + c*z - numpy.sqrt(rootterm))/(a**2 + b**2 + c**2)

def cross_pos(a, b, c, x, y, z, R, rootterm):
	return (a*x + b*y + c*z + numpy.sqrt(rootterm))/(a**2 + b**2 + c**2)

def sample_nh(xx, yy, zz, RR, v, mindistances, conversion):
	a = v[:,0].reshape((1, -1))
	b = v[:,1].reshape((1, -1))
	c = v[:,2].reshape((1, -1))
	root = rootterm(a, b, c, xx, yy, zz, RR)
	mask = root >= 0
	print('collisions: %.2f' % (mask.sum(axis=0) * 1.).mean())
	sys.stdout.flush()
	sola = cross_pos(a, b, c, xx, yy, zz, RR, root)
	solb = cross_neg(a, b, c, xx, yy, zz, RR, root)
	#print mask.shape, sola.shape, solb.shape
	lowr = numpy.min([sola, solb], axis=0)
	highr= numpy.max([sola, solb], axis=0)
	assert lowr.shape == sola.shape
	assert highr.shape == solb.shape
	
	seqs = []
	
	for i, mindistance in enumerate(mindistances):
		lowr[lowr < mindistance] = mindistance
		highr[highr < mindistance] = mindistance
		lengths = numpy.where(mask, numpy.abs(highr - lowr), 0)
		NH = lengths * conversion
		NHtotal = NH.sum(axis=0) + 1e18
		seqs.append(NHtotal.tolist())
		#print '** NH=%.1f obscfrac=%.1f%% CTfrac=%.1f%%' % (numpy.log10(NHtotal).mean(), 
		#	100 * (NHtotal>1e22).mean(), 100 * (NHtotal>1e24).mean())
	return seqs

def nh_dist(x, y, z, R, density, mindistances=[0], ):
	"""
	sample the NH from center to random directions
	NH is integrated along the LOS, starting at mindistance [in ckpc/h]
	x, rho, m are the positions, densities and masses of gas clumps,
	which will be assumed to be constant-density spheres
	"""
	# number of H atoms per cm^3
	# assuming everything is in H, which will be a over-estimate of NH
	Msolar_per_kpc3_in_per_cm3 = 4.04e-8
	kpc_in_cm = 3.086e21
	Hdensity = density * Msolar_per_kpc3_in_per_cm3
	conversion = kpc_in_cm * Hdensity.reshape((-1,1))
	
	xx = x.reshape((-1, 1))
	yy = y.reshape((-1, 1))
	zz = z.reshape((-1, 1))
	RR = R.reshape((-1, 1))
	numpy.random.seed(1)

	vall = numpy.random.normal(size=(nsamples, 3))
	assert vall.sum(axis=1).shape == (nsamples,), v.sum(axis=1).shape
	vall /= ((vall**2).sum(axis=1)**0.5).reshape((-1, 1)) # normalize
	assert vall.shape == (nsamples, 3), vall.shape
	
	N = 1
	vs = [vall[i*N:(i+1)*N] for i in range(nsamples / N)]
	allseqs = [sample_nh(xx, yy, zz, RR, v, mindistances, conversion) for v in vs]
	seqs = [[] for mindistance in mindistances]
	
	for chunkseqs in allseqs:
		for i in range(len(mindistances)):
			seqs[i] += chunkseqs[i]
	for i in range(len(mindistances)):
		NHtotal = numpy.array(seqs[i])
		assert NHtotal.shape == (nsamples,), NHtotal.shape
		print('** NH=%.1f obscfrac=%.1f%% CTfrac=%.1f%%' % (numpy.log10(NHtotal).mean(), 
			100 * (NHtotal>1e22).mean(), 100 * (NHtotal>1e24).mean()))
	NHtotal = numpy.array(seqs)
	return NHtotal


def test():
	data_orig = numpy.load(os.path.join(os.path.dirname(__file__), 'ray_example.npz'))
	data = dict([(k, data_orig[k].astype(numpy.float64)) for k in data_orig.keys()])
	x = data['x']
	y = data['y']
	z = data['z']
	R = data['R']
	rho = data['conversion'][:,0]
	
	vall = data['v'] #[:1,:]
	
	a, b, c = numpy.copy(vall[:,0]), numpy.copy(vall[:,1]), numpy.copy(vall[:,2])
	mindistances = numpy.copy(data['mindistances'])
	
	for i in range(1):
		print('running...')
		result = sphere_raytrace(x, y, z, R, rho, a, b, c, mindistances)
	print('done.')
	print(result)
	assert result.shape == data['NHtotal'].shape, (result.shape, data['NHtotal'].shape)
	print('result:', numpy.log10(result))
	print('reference:', numpy.log10(data['NHtotal']))
	print('absdiff:', numpy.max(numpy.abs(result - data['NHtotal'])))
	print('reldiff:', numpy.max(numpy.abs((result - data['NHtotal'])/data['NHtotal'])))
	assert numpy.allclose(result, data['NHtotal'], rtol=0.0001, atol=1e19)
	

if __name__ == '__main__':
	test()




import numpy
from setuptools import Extension
from Cython.Build import cythonize

from setuptools import setup

with open('README.rst', encoding="utf-8") as readme_file:
    readme = readme_file.read()

compile_args = ['-std=c99', '-lm', '-Wall', '-Wextra', '-Wno-unused-function', '-DNPY_NO_DEPRECATED_API']

setup(
    author='Johannes Buchner',
    packages=['lightrayrider'],
    ext_modules = cythonize([
        Extension(
            "lightrayrider.parallel",
            ["lightrayrider/parallel.pyx"],
            include_dirs=[numpy.get_include()],
            extra_compile_args=compile_args + ["-fopenmp", "-DPARALLEL=1"],
            extra_link_args=compile_args + ["-fopenmp", "-DPARALLEL=1"],
        ),
        Extension(
            "lightrayrider.raytrace",
            ["lightrayrider/raytrace.pyx"],
            include_dirs=[numpy.get_include()],
            extra_compile_args=compile_args,
            extra_link_args=compile_args,
        )
    ]),
    long_description=readme,
)



"""
LightRayRider is a tiny fast library to compute column densities along a ray.

Supported geometries:

* Uniform density grids.
* Spheres of varying size, density and position.
* Co-centred cones of varying opening angles and density.
"""
import os

__version__ = '2.0.2'

print(
    """
You are using the LightRayRider library, which provides optimized calls for
photon propagation and column density computations.
Please cite: Buchner & Bauer (2017), """
    "http://adsabs.harvard.edu/abs/2017MNRAS.465.4348B")

if int(os.environ.get('OMP_NUM_THREADS', '1')) > 1:
    from . import parallel as raytrace
    print("Parallelisation enabled.\n")
else:
    from . import raytrace
    print("Parallelisation disabled (use OMP_NUM_THREADS to enable).\n")

sphere_raytrace = raytrace.py_sphere_raytrace
sphere_raytrace_count_between = raytrace.py_sphere_raytrace_count_between
grid_raytrace = raytrace.py_grid_raytrace
voronoi_raytrace = raytrace.py_voronoi_raytrace
sphere_sphere_collisions = raytrace.py_sphere_sphere_collisions
sphere_raytrace_finite = raytrace.py_sphere_raytrace_finite
cone_raytrace_finite = raytrace.py_cone_raytrace_finite
grid_raytrace_finite = raytrace.py_grid_raytrace_finite
